/* 
   DIPTOOLS VER 2.0
*/

#ifndef DT_IMAGE_HH
#define DT_IMAGE_HH

#include <cstdlib>
#include <string>

#include <Magick++.h>

#include "dtImage_aux.hh"

//! A namespace to prevent global namespace pollution
namespace diptools
{

  //! The primary image class
  class dtImage:public Magick::Image
  {
  protected:

    /*! The image data is is stored as a matrix or 2D array.
     *  The array will store image intensity values. 
     *  For now we are working with grayscale images so dtImage_data_t 
     *  is just another name for double. 
     *  The intensity values should range from 0.0 to 1.0.
     *
     *  The array is like your work area - all operations are to be done on this
     *  array and not directly on the image itself.
     */
    array< dtImage_data_t > data;

    /*
     * This is a 2D array which will contain the 
     * fft transform coefficients of the image
     */
    fft_arr_t fft_data;

  public:

    //---------------------------------------------------------------------
    // Constructors and Destructors
    //---------------------------------------------------------------------

    /*! Default Constructor
     *
     * Does not do anything. Just calls the Image Magick Constructor.
     * Use the `read' method to load an image.
     * DO NOT IMPLEMENT THIS AGAIN.
     */
    dtImage():Magick::Image() { }

    /*! Copy Constructor
     *
     * This copies the data of image given as the parameter to form a new
     * image. The ImageMagick Image class has its own copy constructor.
     * DO NOT IMPLEMENT THIS AGAIN.
     */
    dtImage(const dtImage& _img):Magick::Image()
    {
      data = _img.data;
      fft_data = _img.fft_data;
    }

    /*! Destructor
     *
     * Free the image data arrays.
     * Calls Magick::~Image automatically
     * DO NOT IMPLEMENT THIS AGAIN.
     */
    ~dtImage()
      {
	data.free();
	fft_data.free();
      }

    //----------------------------------------------------------------------
    // Read/Write/Display Methods
    //----------------------------------------------------------------------

    /*!
     *  Use the ImageMagick `read' method to read in the image.
     *  Use `image2array' to copy the pixel values from the image
     *  to the data array.
     */
    void read(const std::string &image_filename);
    
    /*! 
     * Use the ImageMagick `write' method to read in the image.
     * First copy back the data array values back to the image using `array2image'
     * and then write the image.
     */
    void write(const std::string &image_filename);

    /*! 
     * Use the ImageMagick `display' method to show the image.
     * First copy back the data array values back to the image using `array2image'
     * and then show the image.
     */
    void display(void);

    //----------------------------------------------------------------------
    // Array Methods
    //----------------------------------------------------------------------

    /*! 
     * Copy the values from the image to the array.
     */
    void image2array(void);

    /*! 
     * Copy the values back from the array to the image.
     */
    void array2image(void);

    /*****************************************************************************************
     * All functions will modify the existing image data array i.e. the methods will
     * modify the data array of the Image object it is called with.
     *****************************************************************************************/
     

    //-------------------------------------------------------------------------
    // Transfer functions
    //-------------------------------------------------------------------------

    /*
     * This method takes as input a fucntion pointer to a transfer function, and a pointer
     * to the argument for the transfer function.
     * Apply this transfer fucntion to every value of your data array and use it to
     * implement Image inversion, contrast stretch, gamma correction and 
     * dynamic range compression.
     * For transfer function which take multiple arguments use the arg pointer to
     * point to a list of arguments or a structure of arguments
     */
    void intensity_transform(void (*transfer_function)(void *), void *arg);
    
    //-------------------------------------------------------------------------
    // Histogram Operations
    //-------------------------------------------------------------------------

    //! Returns the histogram of the image
    histogram_t get_histogram(void);

    //! Global Histogram Equilization 
    void equalize(void);

    //! Local Histogram Equilization 
    void local_equalize(unsigned int mask_size);

    //! Histogram Specification
    //! This will actually perform histgram specification from the given histogram.
    void set_histogram(const histogram_t& histogram);

    //--------------------------------------------------------------------------
    // Convolution Function
    //--------------------------------------------------------------------------

    /*! 
     *  Convolve the mask with the data array.
     *  Use this function to implement Neighbourhood Averaging filter,
     *  Highpass filter and Highboost filter.
     */
    void convolve(const mask_t &mask);

    //--------------------------------------------------------------------------
    // Median Filter
    //--------------------------------------------------------------------------
    //! Median Filter function
    void median(unsigned int mask_size);
    
    //--------------------------------------------------------------------------
    // Halftoning
    //--------------------------------------------------------------------------

    //! Halftoning function - converts the image to a halftone image
    //! dither_matrix is the pseudo-random dither matrix
    void halftone(const mask_t &dither_matrix);

        
    //-------------------------------------------------------------------------
    // FFT/IFFT Routines
    //-------------------------------------------------------------------------

    //! Computes the FFT of the terms in the 1D array 'arr'. Length of 'arr' is N.
    //! Returns the array of FFT coefficients
    fft_coeff_t* fft_1d(const fft_coeff_t* arr, const unsigned int N);

    //! Computes the FFT of the terms in the 1D array 'arr'. Length of 'arr' is N.
    //! Returns the array of FFT coefficients
    fft_coeff_t* ifft_1d(const fft_coeff_t* arr, const unsigned int N);

    //! Computes the FFT of the terms in the 2D array 'data'.
    //! and stores the results in 'fft_data'
    void fft_2d(void);

    //! Computes the inverse FFT of the terms in the 2D array 'fft_data'.
    //! and stores the results back in 'data'
    void ifft_2d(void);

    //! Generates the image of the power spectrum of the fourier coefficients in 'fft_data'
    dtImage& power_spectrum_image(void);


    //----------------------------------------------------------------------------
    // Frequency Domain Filters
    //----------------------------------------------------------------------------

    //! Applies the Ideal Lowpass filter to the image
    //! Image is in the frequency domain and the filter modifies 'fft_data'
    void lowpass(double Do); 

    //! Applies the Ideal Highpass filter to the image
    //! Image is in the frequency domain and the filter modifies 'fft_data'
    void highpass(double Do); 

    //! Applies the Butterworth Lowpass filter to the image
    //! Image is in the frequency domain and the filter modifies 'fft_data'
    void butterworth_lowpass(double Do, double n); 

    //! Applies the Butterworth Highpass filter to the image
    //! Image is in the frequency domain and the filter modifies 'fft_data'
    void butterworth_highpass(double Do, double n); 

    //! Applies the Homomorphic Gaussian filter to the image
    //! Image is in the frequency domain and the filter modifies 'fft_data'
    void homomorphic_gaussian(double gammaL, double gammaH);

  };



}

#endif
