/* 
   DIPTOOLS VER 2.0

   DO NOT REIMPLEMENT ANY FUNCTION FROM THIS FILE
*/

#ifndef DT_IMAGE_AUX_HH
#define DT_IMAGE_AUX_HH

#include <complex>

//! A namespace to prevent global namespace pollution
namespace diptools
{

  //! The type of data in the image
  typedef double dtImage_data_t ;

  //! The histogram data type
  /*! This will be an array on number of pixels for each intensity value
   *  Intensity values have to be discritized here from 0 - 255 for 8 bit
   *  grayscale images.
   */  
  typedef double* histogram_t;

  //----------------------------------------------------------------------

  //! A templatized array class
  template <class arr_t>class array
  {

  protected:
    arr_t **data;
    unsigned int rows, cols;
    
  public:
    
    //! Default Constructor
    // If this constructor is used to create an array use `resize'
    // to allocate space before using the array
    array()
    {
      rows=0; cols=0; data=NULL;
    }

    //! Constructor
    array(unsigned int inrows, unsigned int incols)
    {
      rows = inrows;
      cols  = incols;
      data = new(arr_t *)[rows];
      for(unsigned int r = 0; r< rows; r++)
	data[r] = new arr_t[cols];
    }

    //! Copy Constructor
    array(const array& arr)
    {
      if (&arr != this)
	{
	  resize(arr.rows, arr.cols);
	  
	  for(unsigned int r = 0; r< rows; r++)
	    for(unsigned int c = 0; c< cols; c++)
	      data[r][c] = arr.data[r][c];
	}
    }

    //! Destructor
    ~array()
    {
      free();
    }

    //! Frees space allocated to the array
    void free(void)
    {
      if (data != NULL)
	{
	  for(unsigned int r = 0; r< rows; r++)
	    delete [](data[r]);
	  
	  delete []data;
	  data = NULL;
	}
      rows = 0; cols = 0;
    }

    //! Resizes the array
    //! Previous data stored in the array is wiped clean
    void resize(unsigned int inrows, unsigned int incols)
    {
      free();

      rows = inrows;
      cols  = incols;
      data = new(arr_t *)[rows];
      for(unsigned int r = 0; r< rows; r++)
	data[r] = new arr_t[cols];
    }

    /*! Overloaded assignment operator
     *  Basically functions similarly to the copy constructor
     */
    array& operator= (const array& arr)
    {
      if (&arr != this)
	{
	  resize(arr.rows, arr.cols);
	  
	  for(unsigned int r = 0; r< rows; r++)
	    for(unsigned int c = 0; c< cols; c++)
	      data[r][c] = arr.data[r][c];
	}
      return *this;
    }

    //! Const access to array members
    //! Can be used to get the array data values
    const arr_t& operator() (const unsigned int i, const unsigned int j) const
    { return data[i][j]; }

    //! Non const access to array members
    //! Can be used to set the array data values
    arr_t& operator() (const unsigned int i, const unsigned int j) 
    { return data[i][j]; }

    //! Returns number of rows of the array
    unsigned int get_rows(void) const { return rows; }

    //! Returns number of cols of the array
    unsigned int get_cols(void) const { return cols; }
  };

  //------------------------------------------------------------------------------

  //! A mask datatype - to be used with spatial filter functions
  class mask_t : public array< double > 
  {
  private:

    //! Mask size
    unsigned int size;

  public:

    //! Constructor
    mask_t(unsigned int _size):array< double >(_size, _size)
    { size = _size; }

    //! Copy Constructor
    mask_t(const mask_t& _mask):array< double >( _mask )
    { size = _mask.size; }

    //! Desturctor
    //! Calls destructor of array automatically
    ~mask_t()
    { }
    
    //! Returns the mask size
    unsigned int get_size(void) const { return size; }
  };

  //----------------------------------------------------------------------

  //! FFT helper types
  typedef std::complex< double > fft_coeff_t;

  //! A 2D fft coefficients array datatype - to be used with frequency filter functions
  class fft_arr_t : public array< fft_coeff_t > 
  {
  public:

    //! Constructor
    fft_arr_t():array< fft_coeff_t >()
    { }

    //! Constructor
    fft_arr_t(unsigned int _rows, unsigned int _cols):array< fft_coeff_t >(_rows, _cols)
    { }

    //! Copy Constructor
    fft_arr_t(const fft_arr_t& _fftarr):array< fft_coeff_t >( _fftarr )
    { }

    //! Desturctor
    //! Calls destructor of array automatically
    ~fft_arr_t()
    { }
  };

  //----------------------------------------------------------------------

}


#endif
